from typing import Optional, Protocol, TypeVar


A = TypeVar("A")
I_contra = TypeVar("I_contra", contravariant=True)


class CollectionOrientedRepository(Protocol[A, I_contra]):
    async def add(self, _aggregate: A) -> None:
        ...

    async def get(self, _id: I_contra) -> Optional[A]:
        ...

    async def delete(self, _aggregate: A) -> None:
        ...


class PersistenceOrientedRepository(Protocol[A, I_contra]):
    async def save(self, _aggregate: A) -> None:
        ...

    async def get(self, _id: I_contra) -> Optional[A]:
        ...

    async def delete(self, _aggregate: A) -> None:
        ...


I_co = TypeVar("I_co", covariant=True)


class GenerateIdentifier(Protocol[I_co]):
    def next_identity(self) -> I_co:
        ...



from contextvars import ContextVar, Token
from dataclasses import field
from types import TracebackType
from typing import Any, Optional, ContextManager
from typing_extensions import Self


_messages_context_var: ContextVar[set[Any]] = ContextVar("messages")


class MessageCatcher(ContextManager["MessageCatcher"]):

    _token: Token[set[Any]] = field(init=False)
    issued_messages: set[Any] = field(init=False)

    def __enter__(self) -> Self:
        self._token = _messages_context_var.set(set())
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        self.issued_messages = _messages_context_var.get()
        _messages_context_var.reset(self._token)


def issue(message: Any):
    _messages_context_var.get().add(message)


def get_issued_messages():
    return _messages_context_var.get()



# type: ignore
from .email_sender import EmailSender
from .outbox import Outbox
from .repository import (
    CollectionOrientedRepository,
    PersistenceOrientedRepository,
    GenerateIdentifier,
)
from .unit_of_work import UnitOfWork



import asyncio
from typing import (
    Any,
    Awaitable,
    Callable,
    Literal,
    Optional,
    TypeVar,
    overload,
)

from pydamain.domain.messages import Message

from .handler import Handler
from .message_catcher import MessageCatcher

T = TypeVar("T")
R = TypeVar("R")

PreHook = Callable[[T, Handler[T, R]], Awaitable[None]]
PostHook = Callable[[T, Handler[T, R], R], Awaitable[None]]
ExceptionHook = Callable[[T, Handler[T, R], Exception], Awaitable[None]]


async def handle(
    message: T,
    handler: Handler[T, Any],
    deps: dict[str, Any],
    pre_hook: Optional[PreHook[T, Any]] = None,
    post_hook: Optional[PostHook[T, Any]] = None,
    exception_hook: Optional[ExceptionHook[T, Any]] = None,
):
    try:
        if pre_hook:
            await pre_hook(message, handler)
        result = await handler(message, **deps)
        if post_hook:
            await post_hook(message, handler, result)
        return result
    except Exception as e:
        if exception_hook:
            await exception_hook(message, handler, e)
        raise e


async def handle_parallel(
    message: T,
    handlers: tuple[Handler[T, Any], ...],
    deps: dict[str, Any],
    pre_hook: Optional[PreHook[T, Any]] = None,
    post_hook: Optional[PostHook[T, Any]] = None,
    exception_hook: Optional[ExceptionHook[T, Any]] = None,
):
    coros = (
        handle(message, handler, deps, pre_hook, post_hook, exception_hook)
        for handler in handlers
    )
    return await asyncio.gather(*coros, return_exceptions=True)


M = TypeVar("M", bound=Message)
Handlers = tuple[Handler[T, R], ...]


class MessageBus:
    def __init__(
        self,
        *,
        deps: dict[str, Any],
        pre_hook: Optional[PreHook[Message, Any]] = None,
        post_hook: Optional[PostHook[Message, Any]] = None,
        exception_hook: Optional[ExceptionHook[Message, Any]] = None,
    ) -> None:
        self._deps: dict[str, Any] = deps
        self._handler_map: dict[
            type[Message], Handler[Any, Any] | Handlers[Any, Any]
        ] = {}
        self._pre_hook = pre_hook
        self._post_hook = post_hook
        self._exception_hook = exception_hook

    def register(
        self,
        message_type: type[M],
        handler: Handler[M, Any] | Handlers[M, Any],
    ):
        self._handler_map[message_type] = handler

    @overload
    async def dispatch(self, message: Message) -> Any:
        ...

    @overload
    async def dispatch(
        self, message: Message, return_hooked_task: Literal[True] = True
    ) -> tuple[Any, asyncio.Future[list[Any]]]:
        ...

    async def dispatch(self, message: Message, return_hooked_task: bool = False):
        result, hooked = await asyncio.create_task(self._dispatch(message))
        coros = (self._dispatch(msg) for msg in hooked)
        hooked_task = asyncio.gather(*coros, return_exceptions=True)
        if return_hooked_task:
            return result, hooked_task
        await hooked_task
        return result

    async def _dispatch(self, message: Message):
        handler = self._handler_map[type(message)]
        with MessageCatcher() as message_catcher:
            if isinstance(handler, tuple):
                result = await handle_parallel(
                    message,
                    handler,
                    self._deps,
                    self._pre_hook,
                    self._post_hook,
                    self._exception_hook,
                )
            else:
                result = await handle(
                    message,
                    handler,
                    self._deps,
                    self._pre_hook,
                    self._post_hook,
                    self._exception_hook,
                )
        return result, message_catcher.issued_messages



from typing import Iterable, Protocol, TypeVar


E = TypeVar("E")


class Outbox(Protocol[E]):
    async def all(self) -> Iterable[E]:
        ...

    async def put(self, _envelope: E) -> None:
        ...

    async def delete(self, _envelope: E) -> None:
        ...
